@@ -1,6 +1,6 @@
 /*
     PowerDNS Versatile Database Driven Nameserver
-    Copyright (C) 2002-2003  PowerDNS.COM BV
+    Copyright (C) 2002-2005  PowerDNS.COM BV
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License version 2 as 
@@ -15,7 +15,7 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
-// $Id: bindbackend2.cc,v 1.6 2003/10/04 14:15:46 ahu Exp $ 
+
 #include <errno.h>
 #include <string>
 #include <map>
@@ -37,7 +37,6 @@ using namespace std;
 #include "bindparser.hh"
 #include "logger.hh"
 #include "arguments.hh"
-#include "huffman.hh"
 #include "qtype.hh"
 #include "misc.hh"
 #include "dynlistener.hh"
@@ -46,7 +45,9 @@ using namespace std;
 
 /** new scheme of things:
     we have zone-id map
-    a zone-id has a vector of DNSResourceRecords */
+    a zone-id has a vector of DNSResourceRecords 
+    on start of query, we find the best zone to answer from
+*/
 
 map<string,int> Bind2Backend::s_name_id_map;
 map<u_int32_t,BB2DomainInfo* > Bind2Backend::s_id_zone_map;
@@ -114,7 +115,6 @@ void Bind2Backend::setFresh(u_int32_t domain_id)
 
 bool Bind2Backend::startTransaction(const string &qname, int id)
 {
-
   BB2DomainInfo &bbd=*s_id_zone_map[d_transaction_id=id];
   d_transaction_tmpname=bbd.d_filename+"."+itoa(random());
   d_of=new ofstream(d_transaction_tmpname.c_str());
@@ -170,7 +170,7 @@ bool Bind2Backend::feedRecord(const DNSResourceRecord &r)
 
   string content=r.content;
 
-  // SOA needs stripping too! XXX FIXME
+  // SOA needs stripping too! XXX FIXME - also, this should not be here I think
   switch(r.qtype.getCode()) {
   case QType::TXT:
     *d_of<<qname<<"\t"<<r.ttl<<"\t"<<r.qtype.getName()<<"\t\""<<r.content<<"\""<<endl;
@@ -217,7 +217,6 @@ void Bind2Backend::getUpdatedMasters(vector<DomainInfo> *changedDomains)
     else
       if(soadata.serial!=i->second->d_lastnotified)
 	changedDomains->push_back(di);
-    
   }
 }
 
@@ -282,7 +281,7 @@ static string canonic(string ret)
   for(i=ret.begin();
       i!=ret.end();
       ++i)
-    ; // *i=*i; //tolower(*i);
+    *i=tolower(*i);
 
 
   if(*(i-1)=='.')
@@ -290,18 +289,45 @@ static string canonic(string ret)
   return ret;
 }
 
+
+set<string> contents;
+
 map<unsigned int, BB2DomainInfo*> nbbds;
 /** This function adds a record to a domain with a certain id. */
 void Bind2Backend::insert(int id, const string &qnameu, const string &qtype, const string &content, int ttl=300, int prio=25)
 {
-  DNSResourceRecord rr;
-  rr.domain_id=id;
-  rr.qname=canonic(qnameu);
-  rr.qtype=qtype;
-  rr.content=content;
-  rr.ttl=ttl;
-  rr.priority=prio;
-  nbbds[id]->d_records->push_back(rr);
+  Bind2DNSRecord bdr;
+
+  //  rr.domain_id=id;
+  BB2DomainInfo* bb2=nbbds[id];
+
+  vector<Bind2DNSRecord> &records=*bb2->d_records;
+
+  bdr.qname=toLower(canonic(qnameu));
+  if(bdr.qname==bb2->d_name)
+    bdr.qname.clear();
+  else if(bdr.qname.length() > bb2->d_name.length())
+    bdr.qname.resize(bdr.qname.length() - (bb2->d_name.length() + 1));
+  else
+    throw AhuException("Trying to insert non-zone data, name='"+bdr.qname+"', zone='"+nbbds[id]->d_name+"'");
+
+  bdr.qname.swap(bdr.qname);
+
+  if(!records.empty() && bdr.qname==(records.end()-1)->qname)
+    bdr.qname=(records.end()-1)->qname;
+
+  bdr.qtype=QType(qtype.c_str()).getCode();
+  bdr.content=canonic(content); // I think this is wrong, the zoneparser should not come up with . terminated stuff XXX FIXME
+  set<string>::const_iterator i=contents.find(bdr.content);
+  if(i!=contents.end())
+   bdr.content=*i;
+  else {
+    contents.insert(bdr.content);
+  }
+
+  bdr.ttl=ttl;
+
+  records.push_back(bdr);
 }
 
 
@@ -383,8 +409,6 @@ static void callback(unsigned int domain_id, const string &domain, const string
   us->insert(domain_id,domain,qtype,content,ttl,prio);
 }
 
-
-
 Bind2Backend::Bind2Backend(const string &suffix)
 {
 #if __GNUC__ >= 3
@@ -401,32 +425,7 @@ Bind2Backend::Bind2Backend(const string &suffix)
    
   s_first=0;
   
-  if(mustDo("example-zones")) {
-    insert(0,"www.example.com","A","1.2.3.4");
-    insert(0,"example.com","SOA","ns1.example.com hostmaster.example.com");
-    insert(0,"example.com","NS","ns1.example.com",86400);
-    insert(0,"example.com","NS","ns2.example.com",86400);
-    insert(0,"example.com","MX","mail.example.com",3600,25);
-    insert(0,"example.com","MX","mail1.example.com",3600,25);
-    insert(0,"mail.example.com","A","4.3.2.1");
-    insert(0,"mail1.example.com","A","5.4.3.2");
-    insert(0,"ns1.example.com","A","4.3.2.1");
-    insert(0,"ns2.example.com","A","5.4.3.2");
-      
-    for(int i=0;i<1000;i++)
-      insert(0,"host-"+itoa(i)+".example.com","A","2.3.4.5");
-
-    s_id_zone_map[0]=new BB2DomainInfo;
-    BB2DomainInfo &bbd=*s_id_zone_map[0];
-    bbd.d_name="example.com";
-    bbd.d_filename="";
-    bbd.d_id=0;
-    s_id_zone_map[0]->d_loaded=true;
-    s_id_zone_map[0]->d_status="parsed into memory at "+nowTime();
-  }
-  
   loadConfig();
-  
 
   extern DynListener *dl;
   us=this;
@@ -457,7 +456,12 @@ void Bind2Backend::loadConfig(string* status)
       L<<Logger::Error<<"Error parsing bind configuration: "<<ae.reason<<endl;
       throw;
     }
-    
+    /*
+    cout<<"sizeof Bind2DNSRecord: "<<sizeof(Bind2DNSRecord)<<endl;
+    cout<<"sizeof string: "<<sizeof(string)<<endl;
+    cout<<"sizeof qtype: "<<sizeof(QType)<<endl;
+    */
+
     ZoneParser ZP;
       
     vector<BindDomainInfo> domains=BP.getDomains();
@@ -488,7 +492,7 @@ void Bind2Backend::loadConfig(string* status)
 	  }
 	if(j==s_id_zone_map.end()) { // entirely new
 	  bbd=new BB2DomainInfo;
-	  bbd->d_records=new vector<DNSResourceRecord>;
+	  bbd->d_records=new vector<Bind2DNSRecord>;
 	  
 	  bbd->d_id=domain_id++;
 	  s_name_id_map[i->name]=bbd->d_id;
@@ -507,10 +511,16 @@ void Bind2Backend::loadConfig(string* status)
 	  
 	  try {
 	    ZP.parse(i->filename,i->name,bbd->d_id); // calls callback for us
+	    L<<Logger::Info<<d_logprefix<<" sorting '"<<i->name<<"'"<<endl;
+	    sort(nbbds[bbd->d_id]->d_records->begin(), nbbds[bbd->d_id]->d_records->end());
 	    nbbds[bbd->d_id]->setCtime();
 	    nbbds[bbd->d_id]->d_loaded=true;          // does this perform locking for us?
 	    nbbds[bbd->d_id]->d_status="parsed into memory at "+nowTime();
-	    
+	    //	    cout<<"Had "<<contents.size()<<" different content fields, "<<nbbds[bbd->d_id]->d_records->size()<<" records, capacity: "<<
+	    //	      nbbds[bbd->d_id]->d_records->capacity()<<endl;
+	    contents.clear();
+	    nbbds[bbd->d_id]->d_records->swap(*nbbds[bbd->d_id]->d_records);
+
 	  }
 	  catch(AhuException &ae) {
 	    ostringstream msg;
@@ -586,7 +596,6 @@ void Bind2Backend::nukeZoneRecords(BB2DomainInfo *bbd)
 /** Must be called with bbd locked already. Will not be unlocked on return, is your own problem.
     Does not throw errors or anything, may update d_status however */
 
-
 void Bind2Backend::queueReload(BB2DomainInfo *bbd)
 {
   nbbds.clear();
@@ -622,75 +631,111 @@ void Bind2Backend::queueReload(BB2DomainInfo *bbd)
   }
 }
 
+bool operator<(const Bind2DNSRecord &a, const string &b)
+{
+  return a.qname < b;
+}
+
+bool operator<(const string &a, const Bind2DNSRecord &b)
+{
+  return a < b.qname;
+}
+
+
 void Bind2Backend::lookup(const QType &qtype,const string &qname, DNSPacket *pkt_p, int zoneId )
 {
-  d_handle=new Bind2Backend::handle;
-  d_handle->d_records=new vector<DNSResourceRecord>; // WRONG
-  //  cout<<"Lookup! for "<<qname<<endl;
-  string domain=qname;
+  d_handle.reset(); // =new Bind2Backend::handle;
+
+  string domain=toLower(qname);
+
+  if(arg().mustDo("query-logging"))
+    L<<"Lookup for '"<<qtype.getName()<<"' of '"<<domain<<"'"<<endl;
+
   while(!s_name_id_map.count(domain) && chopOff(domain));
 
   if(!s_name_id_map.count(domain)) {
-    //    cout<<"no such domain: '"<<qname<<"'"<<endl;
-    d_handle->d_iter=d_handle->d_records->end();  // WRONG
-    d_handle->d_list=false;
-    d_handle->d_bbd=0;
+    d_handle.d_list=false;
+    d_handle.d_bbd=0;
     return;
   }
   unsigned int id=s_name_id_map[domain];
 
-  //  cout<<"domain: '"<<domain<<"', id="<<id<<endl;
+  d_handle.id=id;
   
-
   DLOG(L<<"Bind2Backend constructing handle for search for "<<qtype.getName()<<" for "<<
        qname<<endl);
-
-  d_handle->qname=qname;
-  d_handle->parent=this;
-  d_handle->qtype=qtype;
-  string compressed=toLower(qname);
-  d_handle->d_records=s_id_zone_map[id]->d_records;
-  d_handle->d_bbd=0;
-  if(!d_handle->d_records->empty()) {
+  
+  if(strcasecmp(qname.c_str(),domain.c_str()))
+    d_handle.qname=toLower(qname.substr(0,qname.size()-domain.length()-1)); // strip domain name
+  
+  d_handle.parent=this;
+  d_handle.qtype=qtype;
+  d_handle.domain=domain;
+  d_handle.d_records=s_id_zone_map[id]->d_records; // give it a copy
+  d_handle.d_bbd=0;
+  if(!d_handle.d_records->empty()) {
     BB2DomainInfo& bbd=*s_id_zone_map[id];
     if(!bbd.d_loaded) {
-      delete d_handle;
-      throw DBException("Zone temporarily not available (file missing, or master dead)"); // fuck
+      d_handle.reset();
+      throw DBException("Zone temporarily not available (file missing, or master dead)"); // fsck
     }
 
     if(!bbd.tryRLock()) {
       L<<Logger::Warning<<"Can't get read lock on zone '"<<bbd.d_name<<"'"<<endl;
-      delete d_handle;
-      throw DBException("Temporarily unavailable due to a zone lock"); // fuck
+      d_handle.reset();
+      throw DBException("Temporarily unavailable due to a zone lock"); // fsck
     }
-      
-
+    
     if(!bbd.current()) {
       L<<Logger::Warning<<"Zone '"<<bbd.d_name<<"' ("<<bbd.d_filename<<") needs reloading"<<endl;
       queueReload(&bbd);
     }
-    d_handle->d_bbd=&bbd;
+    d_handle.d_bbd=&bbd;
   }
   else {
     DLOG(L<<"Query with no results"<<endl);
   }
-  d_handle->d_iter=d_handle->d_records->begin();
-  d_handle->d_list=false;
+
+  pair<vector<Bind2DNSRecord>::const_iterator, vector<Bind2DNSRecord>::const_iterator> range;
+
+  //  cout<<"starting equal range for: '"<<d_handle.qname<<"'"<<endl;
+  range=equal_range(d_handle.d_records->begin(), d_handle.d_records->end(), d_handle.qname);
+  
+  if(range.first==range.second) {
+    d_handle.d_bbd=0;
+    d_handle.d_list=false;
+    return;
+  }
+  else {
+    d_handle.d_iter=range.first;
+    d_handle.d_end_iter=range.second;
+  }
+
+  d_handle.d_list=false;
 }
 
 Bind2Backend::handle::handle()
 {
   d_bbd=0;
+  d_records=0;
   count=0;
 }
 
 bool Bind2Backend::get(DNSResourceRecord &r)
 {
-  if(!d_handle->get(r)) {
-    delete d_handle;
-    d_handle=0;
+  if(!d_handle.d_records)
+    return false;
+
+  if(!d_handle.get(r)) {
+    d_handle.reset();
+
+    if(arg().mustDo("query-logging"))
+      L<<"End of answers"<<endl;
+
     return false;
   }
+  if(arg().mustDo("query-logging"))
+    L<<"Returning: '"<<r.qtype.getName()<<"' of '"<<r.qname<<"', content: '"<<r.content<<"'"<<endl;
   return true;
 }
 
@@ -707,11 +752,19 @@ bool Bind2Backend::handle::get_normal(DNSResourceRecord &r)
   DLOG(L << "Bind2Backend get() was called for "<<qtype.getName() << " record  for "<<
        qname<<"- "<<d_records->size()<<" available!"<<endl);
   
-  while(d_iter!=d_records->end() && (d_iter->qname!=qname || !(qtype=="ANY" || (d_iter)->qtype==qtype))) {
+  if(d_iter==d_end_iter) {
+    if(d_bbd) {
+      d_bbd->unlock();
+      d_bbd=0;
+    }
+    return false;
+  }
+
+  while(d_iter!=d_end_iter && !(qtype.getCode()==QType::ANY || (d_iter)->qtype==qtype.getCode())) {
     DLOG(L<<Logger::Warning<<"Skipped "<<qname<<"/"<<QType(d_iter->qtype).getName()<<": '"<<d_iter->content<<"'"<<endl);
     d_iter++;
   }
-  if(d_iter==d_records->end()) { // we've reached the end
+  if(d_iter==d_end_iter) {
     if(d_bbd) {
       d_bbd->unlock();
       d_bbd=0;
@@ -720,13 +773,13 @@ bool Bind2Backend::handle::get_normal(DNSResourceRecord &r)
   }
   DLOG(L << "Bind2Backend get() returning a rr with a "<<QType(d_iter->qtype).getCode()<<endl);
 
-  r.qname=qname; // fill this in
-  
+  r.qname=qname.empty() ? domain : (qname+"."+domain);
+  r.domain_id=id;
   r.content=(d_iter)->content;
-  r.domain_id=(d_iter)->domain_id;
+  //  r.domain_id=(d_iter)->domain_id;
   r.qtype=(d_iter)->qtype;
   r.ttl=(d_iter)->ttl;
-  r.priority=(d_iter)->priority;
+  //  r.priority=(d_iter)->priority;
   d_iter++;
 
   return true;
@@ -734,19 +787,20 @@ bool Bind2Backend::handle::get_normal(DNSResourceRecord &r)
 
 bool Bind2Backend::list(const string &target, int id)
 {
-  // cout<<"List of id "<<id<<" requested"<<endl;
   if(!s_id_zone_map.count(id))
     return false;
 
-  d_handle=new Bind2Backend::handle;
+  d_handle.reset(); // new Bind2Backend::handle;
   DLOG(L<<"Bind2Backend constructing handle for list of "<<id<<endl);
 
-  d_handle->d_qname_iter=s_id_zone_map[id]->d_records->begin();
-  d_handle->d_qname_end=s_id_zone_map[id]->d_records->end();   // iter now points to a vector of pointers to vector<BBResourceRecords>
+  d_handle.d_qname_iter=s_id_zone_map[id]->d_records->begin();
+  d_handle.d_qname_end=s_id_zone_map[id]->d_records->end();   // iter now points to a vector of pointers to vector<BBResourceRecords>
 
-  d_handle->parent=this;
-  d_handle->id=id;
-  d_handle->d_list=true;
+  d_handle.d_records=s_id_zone_map[id]->d_records; // give it a copy --- WHY??? XXX FIXME
+
+  d_handle.parent=this;
+  d_handle.id=id;
+  d_handle.d_list=true;
   return true;
 
 }
@@ -754,7 +808,11 @@ bool Bind2Backend::list(const string &target, int id)
 bool Bind2Backend::handle::get_list(DNSResourceRecord &r)
 {
   if(d_qname_iter!=d_qname_end) {
-    r=*d_qname_iter;
+    r.qname=d_qname_iter->qname.empty() ? domain : (d_qname_iter->qname+"."+domain);
+    r.domain_id=id;
+    r.content=(d_qname_iter)->content;
+    r.qtype=(d_qname_iter)->qtype;
+    r.ttl=(d_qname_iter)->ttl;
     d_qname_iter++;
     return true;
   }
@@ -834,7 +892,7 @@ bool Bind2Backend::createSlaveDomain(const string &ip, const string &domain, con
   
   BB2DomainInfo *bbd = new BB2DomainInfo;
 
-  bbd->d_records = new vector<DNSResourceRecord>;
+  bbd->d_records = new vector<Bind2DNSRecord>;
   bbd->d_name = domain;
   bbd->setCheckInterval(getArgAsNum("check-interval"));
   bbd->d_master = ip;
