@@ -1,11 +1,10 @@
 /*
     PowerDNS Versatile Database Driven Nameserver
-    Copyright (C) 2002  PowerDNS.COM BV
+    Copyright (C) 2005  PowerDNS.COM BV
 
     This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
+    it under the terms of the GNU General Public License version 2 as 
+    published by the Free Software Foundation
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -16,7 +15,7 @@
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
-// $Id: dnspacket.cc,v 1.32 2005/01/11 19:59:00 ahu Exp $
+
 #include "utility.hh"
 #include <cstdio>
 
@@ -50,6 +49,15 @@ string DNSPacket::getString()
   return stringbuffer;
 }
 
+string DNSPacket::getLocal() const
+{
+  struct sockaddr_in sa;
+  int addrlen=sizeof(struct sockaddr_in);
+
+  getsockname(d_socket, (struct sockaddr *)&sa, (socklen_t *)&addrlen);
+  return sockAddrToString(&sa, (socklen_t)addrlen);
+}
+
 
 string DNSPacket::getRemote() const
 {
@@ -80,6 +88,7 @@ DNSPacket::DNSPacket(const DNSPacket &orig)
   d_dt=orig.d_dt;
   d_socklen=orig.d_socklen;
   d_compress=orig.d_compress;
+  d_tcp=orig.d_tcp;
   qtype=orig.qtype;
   qclass=orig.qclass;
   qdomain=orig.qdomain;
@@ -970,29 +979,30 @@ void DNSPacket::wrapup(void)
 
   stable_sort(rrs.begin(),rrs.end(),rrcomp);
 
-  // now shuffle! start out with the ANSWER records  
-  vector<DNSResourceRecord>::iterator first, second;
-  for(first=rrs.begin();first!=rrs.end();++first) 
-    if(first->d_place==DNSResourceRecord::ANSWER && first->qtype.getCode() != QType::CNAME) // CNAME must come first
-      break;
-  for(second=first;second!=rrs.end();++second)
-    if(second->d_place!=DNSResourceRecord::ANSWER)
-      break;
-
-  if(second-first>1)
-    random_shuffle(first,second);
-
-  // now shuffle the additional records
-  for(first=second;first!=rrs.end();++first) 
-    if(first->d_place==DNSResourceRecord::ADDITIONAL && first->qtype.getCode() != QType::CNAME) // CNAME must come first
-      break;
-  for(second=first;second!=rrs.end();++second)
-    if(second->d_place!=DNSResourceRecord::ADDITIONAL)
-      break;
-
-  if(second-first>1)
-    random_shuffle(first,second);
-
+  if(!arg().mustDo("no-shuffle")) {
+    // now shuffle! start out with the ANSWER records  
+    vector<DNSResourceRecord>::iterator first, second;
+    for(first=rrs.begin();first!=rrs.end();++first) 
+      if(first->d_place==DNSResourceRecord::ANSWER && first->qtype.getCode() != QType::CNAME) // CNAME must come first
+	break;
+    for(second=first;second!=rrs.end();++second)
+      if(second->d_place!=DNSResourceRecord::ANSWER)
+	break;
+    
+    if(second-first>1)
+      random_shuffle(first,second);
+    
+    // now shuffle the additional records
+    for(first=second;first!=rrs.end();++first) 
+      if(first->d_place==DNSResourceRecord::ADDITIONAL && first->qtype.getCode() != QType::CNAME) // CNAME must come first
+	break;
+    for(second=first;second!=rrs.end();++second)
+      if(second->d_place!=DNSResourceRecord::ADDITIONAL)
+	break;
+    
+    if(second-first>1)
+      random_shuffle(first,second);
+  }
   d_wrapped=true;
 
 
@@ -1411,6 +1421,7 @@ DNSPacket *DNSPacket::replyPacket() const
   
   r->d_dt=d_dt;
   r->d.qdcount=1;
+  r->d_tcp = d_tcp;
   return r;
 }
 
