@@ -1,11 +1,10 @@
 /*
     PowerDNS Versatile Database Driven Nameserver
-    Copyright (C) 2002  PowerDNS.COM BV
+    Copyright (C) 2002 - 2005 PowerDNS.COM BV
 
     This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
+    it under the terms of the GNU General Public License version 2 as 
+    published by the Free Software Foundation
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -35,7 +34,8 @@
 #include "ahuexception.hh"
 #include "statbag.hh"
 #include "arguments.hh"
-
+#include "sstuff.hh"
+#include "syncres.hh"
 
 LWRes::LWRes()
 {
@@ -55,7 +55,7 @@ LWRes::~LWRes()
 
 //! returns -1 for permanent error, 0 for timeout, 1 for success
 /** Never throws! */
-int LWRes::asyncresolve(const string &ip, const char *domain, int type)
+int LWRes::asyncresolve(const string &ip, const char *domain, int type, bool doTCP)
 {
   DNSPacket p;
   p.setQuestion(Opcode::Query,domain,type);
@@ -69,28 +69,70 @@ int LWRes::asyncresolve(const string &ip, const char *domain, int type)
 
   struct sockaddr_in toaddr;
   struct in_addr inp;
+  Utility::socklen_t addrlen=sizeof(toaddr);
   Utility::inet_aton(ip.c_str(),&inp);
   toaddr.sin_addr.s_addr=inp.s_addr;
 
   toaddr.sin_port=htons(53);
   toaddr.sin_family=AF_INET;
 
-
   int ret;
 
   DTime dt;
   dt.set();
-  if(asendto(p.getData(), p.len, 0, (struct sockaddr*)(&toaddr), sizeof(toaddr),p.d.id)<0) {
-    return -1;
-  }
-    
-  Utility::socklen_t addrlen=sizeof(toaddr);
-  
-  // sleep until we see an answer to this, interface to mtasker
-  
-  ret=arecvfrom(reinterpret_cast<char *>(d_buf), d_bufsize-1,0,(struct sockaddr*)(&toaddr), &addrlen, &d_len, p.d.id);
-  d_usec=dt.udiff();
 
+  if(!doTCP) {
+    if(asendto(p.getData(), p.len, 0, (struct sockaddr*)(&toaddr), sizeof(toaddr),p.d.id)<0) {
+      return -1;
+    }
+  
+    // sleep until we see an answer to this, interface to mtasker
+    
+    ret=arecvfrom(reinterpret_cast<char *>(d_buf), d_bufsize-1,0,(struct sockaddr*)(&toaddr), &addrlen, &d_len, p.d.id);
+  }
+  else {
+    //    cerr<<"do tcp"<<endl;
+    Socket s(InterNetwork, Stream);
+    IPEndpoint ie(ip, 53);
+    s.setNonBlocking();
+    s.connect(ie);
+
+    unsigned int len=htons(p.len);
+    char *lenP=(char*)&len;
+    const char *msgP=p.getData();
+    string packet=string(lenP, lenP+2)+string(msgP, msgP+p.len);
+
+    if(asendtcp(packet, &s) == 0) {
+      //      cerr<<"asendtcp: timeout"<<endl;
+      return 0;
+    }
+    
+    packet.clear();
+    if(arecvtcp(packet,2, &s)==0) {
+      //      cerr<<"arecvtcp: timeout"<<endl;
+      return 0;
+    }
+
+    memcpy(&len, packet.c_str(), 2);
+    len=ntohs(len);
+
+    //    cerr<<"Now reading "<<len<<" bytes"<<endl;
+
+    if(arecvtcp(packet, len, &s)==0) {
+      //      cerr<<"arecvtcp: timeout"<<endl;
+      return 0;
+    }
+    if(len > d_bufsize) {
+      d_bufsize=len;
+      delete[] d_buf;
+      d_buf = new unsigned char[d_bufsize];
+    }
+    memcpy(d_buf, packet.c_str(), len);
+    d_len=len;
+    ret=1;
+  }
+  d_usec=dt.udiff();
+    
   return ret;
 }
 
@@ -103,6 +145,7 @@ LWRes::res_t LWRes::result()
     if(p.parse((char *)d_buf, d_len)<0)
       throw LWResException("resolver: unable to parse packet of "+itoa(d_len)+" bytes");
     d_aabit=p.d.aa;
+    d_tcbit=p.d.tc;
     d_rcode=p.d.rcode;
     return p.getAnswers();
   }
