@@ -1,11 +1,10 @@
 /*
     PowerDNS Versatile Database Driven Nameserver
-    Copyright (C) 2005  PowerDNS.COM BV
+    Copyright (C) 2002-2005  PowerDNS.COM BV
 
     This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
+    it under the terms of the GNU General Public License version 2 as 
+    published by the Free Software Foundation
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -154,7 +153,7 @@ int PacketHandler::doDNSCheckRequest(DNSPacket *p, DNSPacket *r, string &target)
   DNSResourceRecord rr;
 
   if (p->qclass == 3 && p->qtype.getName() == "HINFO") {
-    rr.content = "PowerDNS $Id: packethandler.cc,v 1.27 2005/01/11 19:41:11 ahu Exp $";
+    rr.content = "PowerDNS $Id: packethandler.cc 420 2005-07-02 15:54:09Z ahu $";
     rr.ttl = 5;
     rr.qname=target;
     rr.qtype=13; // hinfo
@@ -174,7 +173,7 @@ int PacketHandler::doVersionRequest(DNSPacket *p, DNSPacket *r, string &target)
   const string mode=arg()["version-string"];
   if(p->qtype.getCode()==QType::TXT && target=="version.bind") {// TXT
     if(mode.empty() || mode=="full") 
-      rr.content="Served by POWERDNS "VERSION" $Id: packethandler.cc,v 1.27 2005/01/11 19:41:11 ahu Exp $";
+      rr.content="Served by POWERDNS "VERSION" $Id: packethandler.cc 420 2005-07-02 15:54:09Z ahu $";
     else if(mode=="anonymous") {
       r->setRcode(RCode::ServFail);
       return 1;
@@ -199,7 +198,7 @@ bool PacketHandler::getAuth(DNSPacket *p, SOAData *sd, const string &target, int
 {
   string subdomain(target);
   do {
-    if( B.getSOA( subdomain, *sd ) ) {
+    if( B.getSOA( subdomain, *sd, p ) ) {
       sd->qname = subdomain;
       *zoneId = sd->domain_id;
       return true;
@@ -286,7 +285,7 @@ int PacketHandler::doAdditionalProcessingAndDropAA(DNSPacket *p, DNSPacket *r)
 	i!=crrs.end();
 	++i) {
       
-      if(i->qtype.getCode()==QType::NS && !B.getSOA(i->qname,sd)) { // drop AA in case of non-SOA-level NS answer
+      if(i->qtype.getCode()==QType::NS && !B.getSOA(i->qname,sd,p)) { // drop AA in case of non-SOA-level NS answer
 	r->d.aa=false;
 	//	i->d_place=DNSResourceRecord::AUTHORITY; // XXX FIXME
       }
@@ -558,7 +557,8 @@ DNSPacket *PacketHandler::question(DNSPacket *p)
     bool found=false;
     
     string target=p->qdomain;
-    
+    bool noCache=false;
+
     if (doDNSCheckRequest(p, r, target))
       goto sendit;
     
@@ -593,7 +593,7 @@ DNSPacket *PacketHandler::question(DNSPacket *p)
 
     if(p->qtype.getCode()==QType::SOA || p->qtype.getCode()==QType::ANY) { // this is special
 
-      if(B.getSOA(target,sd)) {
+      if(B.getSOA(target,sd,p)) {
 	rr.qname=target;
 	rr.qtype=QType::SOA;
 	rr.content=DNSPacket::serializeSOAData(sd);
@@ -667,11 +667,15 @@ DNSPacket *PacketHandler::question(DNSPacket *p)
     else
       weAuth=false;
 
-    if(p->d.rd && d_doRecursion && !weAuth && DP->sendPacket(p)) {
-      delete r;
-      return 0;
-    }
 
+    if(p->d.rd && d_doRecursion && !weAuth) {
+      if(DP->sendPacket(p)) {
+	delete r;
+	return 0;
+      }
+      else noCache=true;
+    }
+    
     string::size_type pos;
     
     DLOG(L<<"Nothing found so far for '"<<target<<"', do we even have authority over this domain?"<<endl);
@@ -757,7 +761,7 @@ DNSPacket *PacketHandler::question(DNSPacket *p)
 	
 	if(!found) {
 	  SOAData sd2;
-	  if(B.getSOA(target,sd2)) // is there a SOA perhaps? (which may not appear in an ANY query)
+	  if(B.getSOA(target,sd2,p)) // is there a SOA perhaps? (which may not appear in an ANY query)
 	    found=true;
 	}
 
@@ -779,16 +783,12 @@ DNSPacket *PacketHandler::question(DNSPacket *p)
     // whatever we've built so far, do additional processing
     
   sendit:;
-
     if(doAdditionalProcessingAndDropAA(p,r)<0)
       return 0;
     
-
-    
-    
-
     r->wrapup(); // needed for inserting in cache
-    PC.insert(p,r); // in the packet cache
+    if(!noCache)
+      PC.insert(p,r); // in the packet cache
   }
   catch(DBException &e) {
     L<<Logger::Error<<"Database module reported condition which prevented lookup - sending out servfail"<<endl;
