@@ -1,11 +1,10 @@
 /*
     PowerDNS Versatile Database Driven Nameserver
-    Copyright (C) 2004  PowerDNS.COM BV
+    Copyright (C) 2003 - 2005  PowerDNS.COM BV
 
     This program is free software; you can redistribute it and/or modify
-    it under the terms of the GNU General Public License as published by
-    the Free Software Foundation; either version 2 of the License, or
-    (at your option) any later version.
+    it under the terms of the GNU General Public License version 2 
+    as published by the Free Software Foundation
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -37,9 +36,16 @@
 #include "syncres.hh"
 #include <fcntl.h>
 #include <fstream>
+#include "sstuff.hh"
+#include <boost/tuple/tuple.hpp>
+#include <boost/tuple/tuple_comparison.hpp>
+#include <boost/shared_array.hpp>
+
+using namespace boost;
+
 #include "recursor_cache.hh"
 
-#ifdef FreeBSD           // see cvstrac ticket #26
+#ifdef __FreeBSD__           // see cvstrac ticket #26
 #include <pthread.h>
 #include <semaphore.h>
 #endif
@@ -48,7 +54,8 @@ MemRecursorCache RC;
 
 string s_programname="pdns_recursor";
 
-#if !WIN32 && !FreeBSD
+#ifndef WIN32
+#ifndef __FreeBSD__
 extern "C" {
   int sem_init(sem_t*, int, unsigned int){return 0;}
   int sem_wait(sem_t*){return 0;}
@@ -59,8 +66,9 @@ extern "C" {
   int pthread_mutex_init(pthread_mutex_t *mutex, const pthread_mutexattr_t *mutexattr){ return 0; }
   int pthread_mutex_lock(pthread_mutex_t *mutex){ return 0; }
   int pthread_mutex_unlock(pthread_mutex_t *mutex) { return 0; }
-
+  int pthread_mutex_destroy(pthread_mutex_t *mutex) { return 0; }
 }
+#endif // __FreeBSD__
 #endif // WIN32
 
 StatBag S;
@@ -70,39 +78,85 @@ ArgvMap &arg()
   return theArg;
 }
 static int d_clientsock;
-static int d_serversock;
-static int d_tcpserversock;
+static vector<int> d_udpserversocks;
+
+typedef vector<int> tcpserversocks_t;
+static tcpserversocks_t s_tcpserversocks;
 
 struct PacketID
 {
-  u_int16_t id;
-  struct sockaddr_in remote;
+  PacketID() : sock(0), inNeeded(0), outPos(0)
+  {}
+
+  u_int16_t id;  // wait for a specific id/remote paie
+  struct sockaddr_in remote;  // this is the remote
+
+  Socket* sock;  // or wait for an event on a TCP fd
+  int inNeeded; // if this is set, we'll read until inNeeded bytes are read
+  string inMSG; // they'll go here
+
+  string outMSG; // the outgoing message that needs to be sent
+  int outPos;    // how far we are along in the outMSG
+
+  bool operator<(const PacketID& b) const
+  {
+    int ourSock= sock ? sock->getHandle() : 0;
+    int bSock = b.sock ? b.sock->getHandle() : 0;
+    return 
+      tie(id, remote.sin_addr.s_addr, remote.sin_port, ourSock) <
+      tie(b.id, b.remote.sin_addr.s_addr, b.remote.sin_port, bSock);
+  }
 };
 
-bool operator<(const PacketID& a, const PacketID& b)
-{
-  if(a.id<b.id)
-    return true;
-
-  if(a.id==b.id) {
-    if(a.remote.sin_addr.s_addr < b.remote.sin_addr.s_addr)
-      return true;
-    if(a.remote.sin_addr.s_addr == b.remote.sin_addr.s_addr)
-      if(a.remote.sin_port < b.remote.sin_port)
-	return true;
-  }
-
-  return false;
-}
+static map<int,PacketID> d_tcpclientreadsocks, d_tcpclientwritesocks;
 
 MTasker<PacketID,string>* MT;
 
+int asendtcp(const string& data, Socket* sock) 
+{
+  PacketID pident;
+  pident.sock=sock;
+  pident.outMSG=data;
+  string packet;
+
+  //  cerr<<"asendtcp called for "<<data.size()<<" bytes"<<endl;
+  d_tcpclientwritesocks[sock->getHandle()]=pident;
+
+  int ret=MT->waitEvent(pident,&packet,1);
+  if(!ret || ret==-1) { // timeout
+    d_tcpclientwritesocks.erase(sock->getHandle());
+  }
+  return ret;
+}
+
+// -1 is error, 0 is timeout, 1 is success
+int arecvtcp(string& data, int len, Socket* sock) 
+{
+  data="";
+  PacketID pident;
+  pident.sock=sock;
+  pident.inNeeded=len;
+
+  // cerr<<"arecvtcp called for "<<len<<" bytes"<<endl;
+  // cerr<<d_tcpclientwritesocks.size()<<" write sockets"<<endl;
+  d_tcpclientreadsocks[sock->getHandle()]=pident;
+
+  int ret=MT->waitEvent(pident,&data,1);
+  if(!ret || ret==-1) { // timeout
+    d_tcpclientreadsocks.erase(sock->getHandle());
+  }
+  return ret;
+}
+
+
 /* these two functions are used by LWRes */
+// -1 is error, > 1 is success
 int asendto(const char *data, int len, int flags, struct sockaddr *toaddr, int addrlen, int id) 
 {
   return sendto(d_clientsock, data, len, flags, toaddr, addrlen);
 }
 
+// -1 is error, 0 is timeout, 1 is success
 int arecvfrom(char *data, int len, int flags, struct sockaddr *toaddr, Utility::socklen_t *addrlen, int *d_len, int id)
 {
   PacketID pident;
@@ -110,19 +164,16 @@ int arecvfrom(char *data, int len, int flags, struct sockaddr *toaddr, Utility::
   memcpy(&pident.remote,toaddr,sizeof(pident.remote));
 
   string packet;
-  if(!MT->waitEvent(pident,&packet,1)) { // timeout
-    return 0; 
+  int ret=MT->waitEvent(pident,&packet,1);
+  if(ret > 0) {
+    *d_len=packet.size();
+    memcpy(data,packet.c_str(),min(len,*d_len));
   }
 
-  *d_len=packet.size();
-  memcpy(data,packet.c_str(),min(len,*d_len));
-
-  return 1;
+  return ret;
 }
 
 
-
-
 static void writePid(void)
 {
   string fname=arg()["socket-dir"]+"/"+s_programname+".pid";
@@ -158,7 +209,7 @@ void primeHints(void)
 
     nsset.insert(nsrr);
   }
-  RC.replace("",QType(QType::NS),nsset);
+  RC.replace("",QType(QType::NS),nsset); // and stuff in the cache
 }
 
 void startDoResolve(void *p)
@@ -168,15 +219,15 @@ void startDoResolve(void *p)
     DNSPacket P=*(DNSPacket *)p;
 
     delete (DNSPacket *)p;
-
+    
     vector<DNSResourceRecord>ret;
     DNSPacket *R=P.replyPacket();
     R->setA(false);
     R->setRA(true);
-
+    //    MT->setTitle("udp question for "+P.qdomain+"|"+P.qtype.getName());
     SyncRes sr;
     if(!quiet)
-      L<<Logger::Error<<"["<<MT->getTid()<<"] question for '"<<P.qdomain<<"|"<<P.qtype.getName()<<"' from "<<P.getRemote()<<endl;
+      L<<Logger::Error<<"["<<MT->getTid()<<"] " << (R->d_tcp ? "TCP " : "") << "question for '"<<P.qdomain<<"|"<<P.qtype.getName()<<"' from "<<P.getRemote()<<endl;
 
     sr.setId(MT->getTid());
     if(!P.d.rd)
@@ -192,20 +243,40 @@ void startDoResolve(void *p)
     }
 
     const char *buffer=R->getData();
-    if(!R->getSocket())
-      sendto(d_serversock,buffer,R->len,0,(struct sockaddr *)(R->remote),R->d_socklen);
+
+    if(!R->d_tcp) {
+      if(R->len > 512) {
+	R->truncate(512);
+      }
+
+      sendto(R->getSocket(),buffer,R->len,0,(struct sockaddr *)(R->remote),R->d_socklen);
+    }
     else {
       char buf[2];
       buf[0]=R->len/256;
       buf[1]=R->len%256;
-      if(write(R->getSocket(),buf,2)!=2 || write(R->getSocket(),buffer,R->len)!=R->len)
+
+      struct iovec iov[2];
+
+      iov[0].iov_base=(void*)buf;    iov[0].iov_len=2;
+      iov[1].iov_base=(void*)buffer; iov[1].iov_len = R->len;
+
+      int ret=writev(R->getSocket(), iov, 2);
+
+      if(ret <= 0 ) 
+	L<<Logger::Error<<"Error writing TCP answer to "<<P.getRemote()<<": "<< (ret ? strerror(errno) : "EOF") <<endl;
+      else if(ret != 2 + R->len)
 	L<<Logger::Error<<"Oops, partial answer sent to "<<P.getRemote()<<" - probably would have trouble receiving our answer anyhow (size="<<R->len<<")"<<endl;
+
+      //      if(write(R->getSocket(),buf,2)!=2 || write(R->getSocket(),buffer,R->len)!=R->len)
+      //  XXX FIXME write this writev fallback otherwise
     }
 
+    //    MT->setTitle("DONE! udp question for "+P.qdomain+"|"+P.qtype.getName());
     if(!quiet) {
       L<<Logger::Error<<"["<<MT->getTid()<<"] answer to "<<(P.d.rd?"":"non-rd ")<<"question '"<<P.qdomain<<"|"<<P.qtype.getName();
       L<<"': "<<ntohs(R->d.ancount)<<" answers, "<<ntohs(R->d.arcount)<<" additional, took "<<sr.d_outqueries<<" packets, "<<
-	sr.d_throttledqueries<<" throttled, "<<sr.d_timeouts<<" timeouts, rcode="<<res<<endl;
+	sr.d_throttledqueries<<" throttled, "<<sr.d_timeouts<<" timeouts, "<<sr.d_tcpoutqueries<<" tcp connections, rcode="<<res<<endl;
     }
     
     sr.d_outqueries ? RC.cacheMisses++ : RC.cacheHits++; 
@@ -230,8 +301,10 @@ void makeClientSocket()
   memset((char *)&sin,0, sizeof(sin));
   
   sin.sin_family = AF_INET;
-  sin.sin_addr.s_addr = INADDR_ANY;
-  
+
+  if(!IpToU32(arg()["query-local-address"], &sin.sin_addr.s_addr))
+    throw AhuException("Unable to resolve local address '"+ arg()["query-local-address"] +"'"); 
+
   int tries=10;
   while(--tries) {
     u_int16_t port=10000+Utility::random()%10000;
@@ -245,74 +318,80 @@ void makeClientSocket()
     throw AhuException("Resolver binding to local socket: "+stringerror());
 
   Utility::setNonBlocking(d_clientsock);
+  L<<Logger::Error<<"Sending UDP queries from "<<inet_ntoa(sin.sin_addr)<<":"<< ntohs(sin.sin_port)  <<endl;
 }
 
-void makeTCPServerSocket()
+void makeTCPServerSockets()
 {
-  d_tcpserversock=socket(AF_INET, SOCK_STREAM,0);
-  if(d_tcpserversock<0) 
-    throw AhuException("Making a server socket for resolver: "+stringerror());
+  vector<string>locals;
+  stringtok(locals,arg()["local-address"]," ,");
+
+  if(locals.empty())
+    throw AhuException("No local address specified");
   
-  struct sockaddr_in sin;
-  memset((char *)&sin,0, sizeof(sin));
+  for(vector<string>::const_iterator i=locals.begin();i!=locals.end();++i) {
+    int fd=socket(AF_INET, SOCK_STREAM,0);
+    if(fd<0) 
+      throw AhuException("Making a server socket for resolver: "+stringerror());
   
-  sin.sin_family = AF_INET;
-
-  if(arg()["local-address"]=="0.0.0.0") {
-    sin.sin_addr.s_addr = INADDR_ANY;
-  }
-  else {
-    if(!IpToU32(arg()["local-address"], &sin.sin_addr.s_addr))
-      throw AhuException("Unable to resolve local address"); 
-  }
-
-  sin.sin_port = htons(arg().asNum("local-port")); 
-
-  int tmp=1;
-  if(setsockopt(d_tcpserversock,SOL_SOCKET,SO_REUSEADDR,(char*)&tmp,sizeof tmp)<0) {
-    L<<Logger::Error<<"Setsockopt failed"<<endl;
-    exit(1);  
-  }
+    struct sockaddr_in sin;
+    memset((char *)&sin,0, sizeof(sin));
     
-  if (bind(d_tcpserversock, (struct sockaddr *)&sin, sizeof(sin))<0) 
-    throw AhuException("TCP Resolver binding to server socket: "+stringerror());
-  
+    sin.sin_family = AF_INET;
+    if(!IpToU32(*i, &sin.sin_addr.s_addr))
+      throw AhuException("Unable to resolve local address '"+ *i +"'"); 
 
-  Utility::setNonBlocking(d_tcpserversock);
-
-  listen(d_tcpserversock, 128);
+    int tmp=1;
+    if(setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,(char*)&tmp,sizeof tmp)<0) {
+      L<<Logger::Error<<"Setsockopt failed for TCP listening socket"<<endl;
+      exit(1);  
+    }
+    
+    sin.sin_port = htons(arg().asNum("local-port")); 
+    
+    if (bind(fd, (struct sockaddr *)&sin, sizeof(sin))<0) 
+      throw AhuException("Binding TCP server socket for "+*i+": "+stringerror());
+    
+    Utility::setNonBlocking(fd);
+    listen(fd, 128);
+    s_tcpserversocks.push_back(fd);
+    L<<Logger::Error<<"Listening for TCP queries on "<<inet_ntoa(sin.sin_addr)<<":"<<arg().asNum("local-port")<<endl;
+  }
 }
 
-void makeServerSocket()
+void makeUDPServerSockets()
 {
-  d_serversock=socket(AF_INET, SOCK_DGRAM,0);
-  if(d_serversock<0) 
-    throw AhuException("Making a server socket for resolver: "+stringerror());
-  
-  struct sockaddr_in sin;
-  memset((char *)&sin,0, sizeof(sin));
-  
-  sin.sin_family = AF_INET;
+  vector<string>locals;
+  stringtok(locals,arg()["local-address"]," ,");
 
+  if(locals.empty())
+    throw AhuException("No local address specified");
+  
   if(arg()["local-address"]=="0.0.0.0") {
     L<<Logger::Warning<<"It is advised to bind to explicit addresses with the --local-address option"<<endl;
-    sin.sin_addr.s_addr = INADDR_ANY;
-  }
-  else {
-    
-    if(!IpToU32(arg()["local-address"], &sin.sin_addr.s_addr))
-      throw AhuException("Unable to resolve local address"); 
-
   }
 
-  sin.sin_port = htons(arg().asNum("local-port")); 
+  for(vector<string>::const_iterator i=locals.begin();i!=locals.end();++i) {
+    int fd=socket(AF_INET, SOCK_DGRAM,0);
+    if(fd<0) 
+      throw AhuException("Making a server socket for resolver: "+stringerror());
+  
+    struct sockaddr_in sin;
+    memset((char *)&sin,0, sizeof(sin));
     
-  if (bind(d_serversock, (struct sockaddr *)&sin, sizeof(sin))<0) 
-    throw AhuException("Resolver binding to server socket: "+stringerror());
-
-  Utility::setNonBlocking(d_serversock);
-
-  L<<Logger::Error<<"Incoming query source port: "<<arg().asNum("local-port")<<endl;
+    sin.sin_family = AF_INET;
+    if(!IpToU32(*i, &sin.sin_addr.s_addr))
+      throw AhuException("Unable to resolve local address '"+ *i +"'"); 
+    
+    sin.sin_port = htons(arg().asNum("local-port")); 
+    
+    if (bind(fd, (struct sockaddr *)&sin, sizeof(sin))<0) 
+      throw AhuException("Resolver binding to server socket for "+*i+": "+stringerror());
+    
+    Utility::setNonBlocking(fd);
+    d_udpserversocks.push_back(fd);
+    L<<Logger::Error<<"Listening for UDP queries on "<<inet_ntoa(sin.sin_addr)<<":"<<arg().asNum("local-port")<<endl;
+  }
 }
 
 
@@ -340,35 +419,42 @@ void usr1Handler(int)
   statsWanted=true;
 }
 
+void usr2Handler(int)
+{
+  SyncRes::setLog(true);
+}
+
+
 
 void doStats(void)
 {
   if(qcounter) {
-    
     L<<Logger::Error<<"stats: "<<qcounter<<" questions, "<<RC.size()<<" cache entries, "<<SyncRes::s_negcache.size()<<" negative entries, "
      <<(int)((RC.cacheHits*100.0)/(RC.cacheHits+RC.cacheMisses))<<"% cache hits";
     L<<Logger::Error<<", outpacket/query ratio "<<(int)(SyncRes::s_outqueries*100.0/SyncRes::s_queries)<<"%";
     L<<Logger::Error<<", "<<(int)(SyncRes::s_throttledqueries*100.0/(SyncRes::s_outqueries+SyncRes::s_throttledqueries))<<"% throttled, "
      <<SyncRes::s_nodelegated<<" no-delegation drops"<<endl;
-    L<<Logger::Error<<"queries running: "<<MT->numProcesses()<<endl;
-    
+    L<<Logger::Error<<"stats: "<<SyncRes::s_tcpoutqueries<<" outgoing tcp connections, "<<MT->numProcesses()<<" queries running, "<<SyncRes::s_outgoingtimeouts<<" outgoing timeouts"<<endl;
   }
+  else if(statsWanted) 
+    L<<Logger::Error<<"stats: no stats yet!"<<endl;
+
   statsWanted=false;
 }
 
 void houseKeeping(void *)
 {
   static time_t last_stat, last_rootupdate, last_prune;
-
-  if(time(0)-last_stat>60) { 
+  time_t now=time(0);
+  if(now - last_prune > 60) { 
     RC.doPrune();
     last_prune=time(0);
   }
-  if(time(0)-last_stat>1800) { 
+  if(now - last_stat>1800) { 
     doStats();
     last_stat=time(0);
   }
-  if(time(0)-last_rootupdate>7200) {
+  if(now -last_rootupdate>7200) {
     SyncRes sr;
     vector<DNSResourceRecord>ret;
 
@@ -376,7 +462,7 @@ void houseKeeping(void *)
     int res=sr.beginResolve("", QType(QType::NS), ret);
     if(!res) {
       L<<Logger::Error<<"Refreshed . records"<<endl;
-      last_rootupdate=time(0);
+      last_rootupdate=now;
     }
     else
       L<<Logger::Error<<"Failed to update . records, RCODE="<<res<<endl;
@@ -391,10 +477,12 @@ struct TCPConnection
   int bytesread;
   struct sockaddr_in remote;
   char data[65535];
+  time_t startTime;
 };
 
 int main(int argc, char **argv) 
 {
+  int ret = EXIT_SUCCESS;
 #ifdef WIN32
     WSADATA wsaData;
     WSAStartup( MAKEWORD( 2, 0 ), &wsaData );
@@ -404,15 +492,23 @@ int main(int argc, char **argv)
     Utility::srandom(time(0));
     arg().set("soa-minimum-ttl","Don't change")="0";
     arg().set("soa-serial-offset","Don't change")="0";
+    arg().set("no-shuffle","Don't change")="off";
     arg().set("aaaa-additional-processing","turn on to do AAAA additional processing (slow)")="off";
     arg().set("local-port","port to listen on")="53";
-    arg().set("local-address","single address to listen on")="0.0.0.0";
+    arg().set("local-address","IP addresses to listen on, separated by spaces or commas")="0.0.0.0";
     arg().set("trace","if we should output heaps of logging")="off";
     arg().set("daemon","Operate as a daemon")="yes";
+    arg().set("chroot","switch to chroot jail")="";
+    arg().set("setgid","If set, change group id to this gid for more security")="";
+    arg().set("setuid","If set, change user id to this uid for more security")="";
     arg().set("quiet","Suppress logging of questions and answers")="off";
     arg().set("config-dir","Location of configuration directory (recursor.conf)")=SYSCONFDIR;
     arg().set("socket-dir","Where the controlsocket will live")=LOCALSTATEDIR;
     arg().set("delegation-only","Which domains we only accept delegations from")="";
+    arg().set("query-local-address","Source IP address for sending queries")="0.0.0.0";
+    arg().set("client-tcp-timeout","Timeout in seconds when talking to TCP clients")="2";
+    arg().set("max-tcp-clients","Maximum number of simultaneous TCP clients")="128";
+
     arg().setCmd("help","Provide a helpful message");
     L.toConsole(Logger::Warning);
     arg().laxParse(argc,argv); // do a lax parse
@@ -435,12 +531,23 @@ int main(int argc, char **argv)
 
     L.setName("pdns_recursor");
 
+    L<<Logger::Warning<<"PowerDNS recursor "<<VERSION<<" (C) 2001-2005 PowerDNS.COM BV ("<<__DATE__", "__TIME__;
+#ifdef __GNUC__
+    L<<", gcc "__VERSION__;
+#endif // add other compilers here
+    L<<") starting up"<<endl;
+
+  L<<Logger::Warning<<"PowerDNS comes with ABSOLUTELY NO WARRANTY. "
+    "This is free software, and you are welcome to redistribute it "
+    "according to the terms of the GPL version 2."<<endl;
+
+
     if(arg().mustDo("trace"))
       SyncRes::setLog(true);
     
     makeClientSocket();
-    makeServerSocket();
-    makeTCPServerSocket();
+    makeUDPServerSockets();
+    makeTCPServerSockets();
         
     MT=new MTasker<PacketID,string>(100000);
 
@@ -456,17 +563,38 @@ int main(int argc, char **argv)
       daemonize();
     }
     signal(SIGUSR1,usr1Handler);
+    signal(SIGUSR2,usr2Handler);
+    signal(SIGPIPE,SIG_IGN);
 
     writePid();
 #endif
 
+    int newgid=0;
+    if(!arg()["setgid"].empty())
+      newgid=Utility::makeGidNumeric(arg()["setgid"]);
+    int newuid=0;
+    if(!arg()["setuid"].empty())
+      newuid=Utility::makeUidNumeric(arg()["setuid"]);
+
+
+    if (!arg()["chroot"].empty()) {
+        if (chroot(arg()["chroot"].c_str())<0) {
+            L<<Logger::Error<<"Unable to chroot to '"+arg()["chroot"]+"': "<<strerror (errno)<<", exiting"<<endl;
+	    exit(1);
+	}
+    }
+
+    Utility::dropPrivs(newuid, newgid);
+
     vector<TCPConnection> tcpconnections;
     counter=0;
+    time_t now=0;
+    unsigned int maxTcpClients=arg().asNum("max-tcp-clients");
     for(;;) {
       while(MT->schedule()); // housekeeping, let threads do their thing
       
       if(!((counter++)%100)) 
-	MT->makeThread(houseKeeping,0);
+	MT->makeThread(houseKeeping,0,"housekeeping");
       if(statsWanted)
 	doStats();
 
@@ -478,19 +606,53 @@ int main(int argc, char **argv)
       tv.tv_sec=0;
       tv.tv_usec=500000;
       
-      fd_set readfds;
+      fd_set readfds, writefds;
       FD_ZERO( &readfds );
+      FD_ZERO( &writefds );
       FD_SET( d_clientsock, &readfds );
-      FD_SET( d_serversock, &readfds );
-      FD_SET( d_tcpserversock, &readfds );
-      int fdmax=max(d_tcpserversock,max(d_clientsock,d_serversock));
-      for(vector<TCPConnection>::const_iterator i=tcpconnections.begin();i!=tcpconnections.end();++i) {
-	FD_SET(i->fd, &readfds);
-	fdmax=max(fdmax,i->fd);
+      int fdmax=d_clientsock;
+
+      if(!tcpconnections.empty())
+	now=time(0);
+
+      vector<TCPConnection> sweeped;
+      int tcpLimit=arg().asNum("client-tcp-timeout");
+      for(vector<TCPConnection>::iterator i=tcpconnections.begin();i!=tcpconnections.end();++i) {
+	if(now < i->startTime + tcpLimit) {
+	  FD_SET(i->fd, &readfds);
+	  fdmax=max(fdmax,i->fd);
+	  sweeped.push_back(*i);
+	}
+	else {
+	  L<<Logger::Error<<"TCP timeout from client "<<inet_ntoa(i->remote.sin_addr)<<endl;
+	  close(i->fd);
+	}
+      }
+      sweeped.swap(tcpconnections);
+
+      for(vector<int>::const_iterator i=d_udpserversocks.begin(); i!=d_udpserversocks.end(); ++i) {
+	FD_SET( *i, &readfds );
+	fdmax=max(fdmax,*i);
+      }
+      if(tcpconnections.size() < maxTcpClients) 
+	for(tcpserversocks_t::const_iterator i=s_tcpserversocks.begin(); i!=s_tcpserversocks.end(); ++i) {
+	  FD_SET(*i, &readfds );
+	  fdmax=max(fdmax,*i);
+	}
+
+      for(map<int,PacketID>::const_iterator i=d_tcpclientreadsocks.begin(); i!=d_tcpclientreadsocks.end(); ++i) {
+	// cerr<<"Adding TCP socket "<<i->first<<" to read select set"<<endl;
+	FD_SET( i->first, &readfds );
+	fdmax=max(fdmax,i->first);
       }
 
-      /* this should listen on a TCP port as well for new connections,  */
-      int selret = select(  fdmax + 1, &readfds, NULL, NULL, &tv );
+      for(map<int,PacketID>::const_iterator i=d_tcpclientwritesocks.begin(); i!=d_tcpclientwritesocks.end(); ++i) {
+	// cerr<<"Adding TCP socket "<<i->first<<" to write select set"<<endl;
+	FD_SET( i->first, &writefds );
+	fdmax=max(fdmax,i->first);
+      }
+
+      int selret = select(  fdmax + 1, &readfds, &writefds, NULL, &tv );
       if(selret<=0) 
 	if (selret == -1 && errno!=EINTR) 
 	  throw AhuException("Select returned: "+stringerror());
@@ -508,7 +670,6 @@ int main(int argc, char **argv)
 	}
 	else { 
 	  if(P.d.qr) {
-
 	    pident.remote=fromaddr;
 	    pident.id=P.d.id;
 	    string packet;
@@ -520,41 +681,105 @@ int main(int argc, char **argv)
 	}
       }
       
-      if(FD_ISSET(d_serversock,&readfds)) { // do we have a new question on udp?
-	d_len=recvfrom(d_serversock, data, sizeof(data), 0, (sockaddr *)&fromaddr, &addrlen);    
-	if(d_len<0) 
-	  continue;
- 	P.setRemote((struct sockaddr *)&fromaddr, addrlen);
-	if(P.parse(data,d_len)<0) {
-	  L<<Logger::Error<<"Unparseable packet from remote client "<<P.getRemote()<<endl;
-	}
-	else { 
-	  if(P.d.qr)
-	    L<<Logger::Error<<"Ignoring answer on server socket!"<<endl;
-	  else {
-	    ++qcounter;
-	    P.setSocket(0);
-	    MT->makeThread(startDoResolve,(void*)new DNSPacket(P));
-
+      for(vector<int>::const_iterator i=d_udpserversocks.begin(); i!=d_udpserversocks.end(); ++i) {
+	if(FD_ISSET(*i,&readfds)) { // do we have a new question on udp?
+	  d_len=recvfrom(*i, data, sizeof(data), 0, (sockaddr *)&fromaddr, &addrlen);    
+	  if(d_len<0) 
+	    continue;
+	  P.setRemote((struct sockaddr *)&fromaddr, addrlen);
+	  if(P.parse(data,d_len)<0) {
+	    L<<Logger::Error<<"Unparseable packet from remote client "<<P.getRemote()<<endl;
+	  }
+	  else { 
+	    if(P.d.qr)
+	      L<<Logger::Error<<"Ignoring answer on server socket!"<<endl;
+	    else {
+	      ++qcounter;
+	      P.setSocket(*i);
+	      P.d_tcp=false;
+	      MT->makeThread(startDoResolve,(void*)new DNSPacket(P), "udp");
+	    }
 	  }
 	}
       }
 
-      if(FD_ISSET(d_tcpserversock,&readfds)) { // do we have a new TCP connection
-	struct sockaddr_in addr;
-	socklen_t addrlen=sizeof(addr);
-	int newsock=accept(d_tcpserversock, (struct sockaddr*)&addr, &addrlen);
-
-	if(newsock>0) {
-	  Utility::setNonBlocking(newsock);
-	  TCPConnection tc;
-	  tc.fd=newsock;
-	  tc.state=TCPConnection::BYTE0;
-	  tc.remote=addr;
-	  tcpconnections.push_back(tc);
+      for(tcpserversocks_t::const_iterator i=s_tcpserversocks.begin(); i!=s_tcpserversocks.end(); ++i) { 
+	if(FD_ISSET(*i ,&readfds)) { // do we have a new TCP connection
+	  struct sockaddr_in addr;
+	  socklen_t addrlen=sizeof(addr);
+	  int newsock=accept(*i, (struct sockaddr*)&addr, &addrlen);
+	  
+	  if(newsock>0) {
+	    Utility::setNonBlocking(newsock);
+	    TCPConnection tc;
+	    tc.fd=newsock;
+	    tc.state=TCPConnection::BYTE0;
+	    tc.remote=addr;
+	    tc.startTime=time(0);
+	    tcpconnections.push_back(tc);
+	  }
 	}
       }
 
+      for(map<int,PacketID>::iterator i=d_tcpclientreadsocks.begin(); i!=d_tcpclientreadsocks.end();) { 
+	bool haveErased=false;
+	if(FD_ISSET(i->first, &readfds)) { // can we receive
+	  shared_array<char> buffer(new char[i->second.inNeeded]);
+
+	  int ret=read(i->first, buffer.get(), min(i->second.inNeeded,200));
+	  // cerr<<"Read returned "<<ret<<endl;
+	  if(ret > 0) {
+	    i->second.inMSG.append(&buffer[0], &buffer[ret]);
+	    i->second.inNeeded-=ret;
+	    if(!i->second.inNeeded) {
+	      // cerr<<"Got entire load of "<<i->second.inMSG.size()<<" bytes"<<endl;
+	      PacketID pid=i->second;
+	      string msg=i->second.inMSG;
+	      
+	      d_tcpclientreadsocks.erase((i++));
+	      haveErased=true;
+	      MT->sendEvent(pid, &msg);   // XXX DODGY
+	    }
+	    else {
+	      // cerr<<"Still have "<<i->second.inNeeded<<" left to go"<<endl;
+	    }
+	  }
+	  else {
+	    //	    cerr<<"when reading ret="<<ret<<endl;
+	    // XXX FIXME I think some stuff needs to happen here - like send an EOF event
+	  }
+	}
+	if(!haveErased)
+	  ++i;
+      }
+
+      for(map<int,PacketID>::iterator i=d_tcpclientwritesocks.begin(); i!=d_tcpclientwritesocks.end(); ) { 
+	bool haveErased=false;
+	if(FD_ISSET(i->first, &writefds)) { // can we send over TCP
+	  // cerr<<"Socket "<<i->first<<" available for writing"<<endl;
+	  int ret=write(i->first, i->second.outMSG.c_str(), i->second.outMSG.size() - i->second.outPos);
+	  if(ret > 0) {
+	    i->second.outPos+=ret;
+	    if(i->second.outPos==i->second.outMSG.size()) {
+	      // cerr<<"Sent out entire load of "<<i->second.outMSG.size()<<" bytes"<<endl;
+	      PacketID pid=i->second;
+	      d_tcpclientwritesocks.erase((i++));
+	      MT->sendEvent(pid, 0);
+	      haveErased=true;
+	      // cerr<<"Sent event too"<<endl;
+	    }
+
+	  }
+	  else { 
+	    //	    cerr<<"ret="<<ret<<" when writing"<<endl;
+	    // XXX FIXME I think some stuff needs to happen here - like send an EOF event
+	  }
+	}
+	if(!haveErased)
+	  ++i;
+      }
+
+
       for(vector<TCPConnection>::iterator i=tcpconnections.begin();i!=tcpconnections.end();++i) {
 	if(FD_ISSET(i->fd, &readfds)) {
 	  if(i->state==TCPConnection::BYTE0) {
@@ -607,12 +832,13 @@ int main(int argc, char **argv)
 	      }
 	      else { 
 		P.setSocket(i->fd);
+		P.d_tcp=true;
 		P.setRemote((struct sockaddr *)&i->remote,sizeof(i->remote));
 		if(P.d.qr)
 		  L<<Logger::Error<<"Ignoring answer on server socket!"<<endl;
 		else {
 		  ++qcounter;
-		  MT->makeThread(startDoResolve,(void*)new DNSPacket(P));
+		  MT->makeThread(startDoResolve,(void*)new DNSPacket(P), "tcp");
 		}
 	      }
 	    }
@@ -623,17 +849,20 @@ int main(int argc, char **argv)
   }
   catch(AhuException &ae) {
     L<<Logger::Error<<"Exception: "<<ae.reason<<endl;
+    ret=EXIT_FAILURE;
   }
   catch(exception &e) {
     L<<Logger::Error<<"STL Exception: "<<e.what()<<endl;
+    ret=EXIT_FAILURE;
   }
   catch(...) {
     L<<Logger::Error<<"any other exception in main: "<<endl;
+    ret=EXIT_FAILURE;
   }
   
 #ifdef WIN32
   WSACleanup();
 #endif // WIN32
 
-  return 0;
+  return ret;
 }
